/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2013  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-expr.c 150 2013-04-11 14:14:53Z acau $
 */


#include "system.h"

int get_seed ( void )

/******************************************************************************/
/*
  Purpose:

    GET_SEED returns a random seed for the random number generator.

  Licensing:

    This code is distributed under the GNU LGPL license. 

  Modified:

    17 November 2004

  Author:

    John Burkardt

  Parameters:

    Output, int GET_SEED, a random seed value.
*/
{
  time_t clock;
      /*int i;*/
  int i4_huge = 2147483647;
  int ihour;
  int imin;
  int isec;
  int seed;
  struct tm *lt;
  time_t tloc;
/*
  If the internal seed is 0, generate a value based on the time.
*/
  clock = time ( &tloc );
  lt = localtime ( &clock );
/*
  Hours is 1, 2, ..., 12.
*/
  ihour = lt->tm_hour;

  if ( 12 < ihour )
  {
    ihour = ihour - 12;
  }
/*
  Move Hours to 0, 1, ..., 11
*/
  ihour = ihour - 1;

  imin = lt->tm_min;

  isec = lt->tm_sec;

  seed = isec + 60 * ( imin + 60 * ihour );
/*
  We want values in [1,43200], not [0,43199].
*/
  seed = seed + 1;
/*
  Remap SEED from [1,43200] to [1,HUGE].
*/
  seed = ( int ) 
    ( ( ( double ) seed )
    * ( ( double ) i4_huge ) / ( 60.0 * 60.0 * 12.0 ) );
/*
  Never use a seed of 0.
*/
  if ( seed == 0 )
  {
    seed = 1;
  }

  return seed;
}

float r4_uniform_01 ( int *seed )

/******************************************************************************/
/*
  Purpose:

    R4_UNIFORM_01 returns a unit pseudorandom R4.

  Discussion:

    This routine implements the recursion

      seed = 16807 * seed mod ( 2**31 - 1 )
      r4_uniform_01 = seed / ( 2**31 - 1 )

    The integer arithmetic never requires more than 32 bits,
    including a sign bit.

    If the initial seed is 12345, then the first three computations are

      Input     Output      R4_UNIFORM_01
      SEED      SEED

         12345   207482415  0.096616
     207482415  1790989824  0.833995
    1790989824  2035175616  0.947702

  Licensing:

    This code is distributed under the GNU LGPL license. 

  Modified:

    16 November 2004

  Author:

    John Burkardt

  Reference:

    Paul Bratley, Bennett Fox, Linus Schrage,
    A Guide to Simulation,
    Second Edition,
    Springer, 1987,
    ISBN: 0387964673,
    LC: QA76.9.C65.B73.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, December 1986, pages 362-376.

    Pierre L'Ecuyer,
    Random Number Generation,
    in Handbook of Simulation,
    edited by Jerry Banks,
    Wiley, 1998,
    ISBN: 0471134031,
    LC: T57.62.H37.

    Peter Lewis, Allen Goodman, James Miller,
    A Pseudo-Random Number Generator for the System/360,
    IBM Systems Journal,
    Volume 8, Number 2, 1969, pages 136-143.

  Parameters:

    Input/output, int *SEED, the "seed" value.  Normally, this
    value should not be 0.  On output, SEED has been updated.

    Output, float R4_UNIFORM_01, a new pseudorandom variate, strictly between
    0 and 1.
*/
{
  int i4_huge = 2147483647;
  int k;
  float r;

  if ( *seed == 0 )
  {
      *seed = get_seed();
  }

  k = *seed / 127773;

  *seed = 16807 * ( *seed - k * 127773 ) - k * 2836;

  if ( *seed < 0 )
  {
    *seed = *seed + i4_huge;
  }
/*
  Although SEED can be represented exactly as a 32 bit integer,
  it generally cannot be represented exactly as a 32 bit real number!
*/
  r = ( float ) ( *seed ) * 4.656612875E-10;

  return r;
}


int i4_uniform_0i ( int *seed )

/******************************************************************************/
/*
  Purpose:

    I4_UNIFORM_0I is a portable random integer generator.

  Formula:

    SEED = SEED * (7**5) mod (2**31 - 1)

  Licensing:

    This code is distributed under the GNU LGPL license. 

  Modified:

    24 September 2005

  Author:

    John Burkardt

  Reference:

    Paul Bratley, Bennett Fox, Linus Schrage,
    A Guide to Simulation,
    Second Edition,
    Springer, 1987,
    ISBN: 0387964673,
    LC: QA76.9.C65.B73.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, December 1986, pages 362-376.

    Pierre L'Ecuyer,
    Random Number Generation,
    in Handbook of Simulation,
    edited by Jerry Banks,
    Wiley, 1998,
    ISBN: 0471134031,
    LC: T57.62.H37.

    Peter Lewis, Allen Goodman, James Miller,
    A Pseudo-Random Number Generator for the System/360,
    IBM Systems Journal,
    Volume 8, Number 2, 1969, pages 136-143.

  Parameters:

    Input, int *SEED, the integer "seed" used to generate
    the output value.  SEED should not be 0.  On output, SEED
    has been updated.

    Output, int I4_UNIFORM_0I, a uniform random value between
    1 and 2**31-1.

  Local parameters:

    IA = 7**5
    IB = 2**15
    IB16 = 2**16
    IP = 2**31-1
*/
{
  int i4_huge = 2147483647;
  int ia = 16807;
  int ib15 = 32768;
  int ib16 = 65536;
  int ip = 2147483647;
  int iprhi;
  int ixhi;
  int k;
  int leftlo;
  int loxa;
  int value;
/*
  Don't let SEED be 0.
*/
  if ( *seed == 0 )
  {
      *seed = get_seed();
  }
/*
  Get the 15 high order bits of SEED2.
*/
  ixhi = *seed / ib16;
/*
  Get the 16 low bits of SEED and form the low product.
*/
  loxa = ( *seed - ixhi * ib16 ) * ia;
/*
  Get the 15 high order bits of the low product.
*/
  leftlo = loxa / ib16;
/*
  Form the 31 highest bits of the full product.
*/
  iprhi = ixhi * ia + leftlo;
/*
  Get overflow past the 31st bit of full product.
*/
  k = iprhi / ib15;
/*
  Assemble all the parts and presubtract IP.  The parentheses are
  essential.
*/
  value = ( ( ( loxa - leftlo * ib16 ) - ip ) 
            + ( iprhi - k * ib15 ) * ib16 ) + k;
/*
  Add IP back in if necessary.
*/
  if ( value < 0 )
  {
    value = value + i4_huge;
  }
  *seed = value;
  return value;
}








/****************************************\
*                                        *
*  statements and expressions for types  *
*                                        *
\****************************************/


/* Get the type of something. */
/* type e */
void reduce_TYPE(e)
NODEPTR e;
{
    NODEPTR loc =reduce_ref  (node_lnk_1(e));

    if (Ready) {
       if ((Ready = mem_defined1 (loc)))
           result (string_node (strsave (print_name (get_mem (loc)))));
    }
}


/* Check whether e is undefined. */
/* _undefined e */
void reduce_iUNDEFINED(e)
NODEPTR e;
{
     NODEPTR loc=reduce_ref (node_lnk_1(e));

     if ((Ready && mem_defined1 (loc)) || ForceOutput) {
         if (Ready || ForceOutput) {
             if (mem_defined1 (loc)){
                 node_mem_isundefined(loc)=0;
                 result(False);
             }
             else{
                 node_mem_isundefined(loc)=1;
                 result(True);
             }
         } 
     }
     
    
    Ready = (Ready &&  mem_defined1 (loc)) || ForceOutput;
}

/**************************************\
*                                      *
*  Simple Boolean, arithmetic, string  *
*  and list expressions.               *
*                                      *
\**************************************/

/*Equality and Non-equality. */
/* e1 = e2, e1 ~= e2 */
void reduce_EQ_val(e)
NODEPTR e;
{
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        e2 = reduce_val (node_lnk_2(e));
        if (Ready) {
            if (node_type(e) == EQ)
                result (equal_val (e1, e2) ? True : False);
            else
                result (equal_val (e1, e2) ? False : True);
        }
    }
}




#define __HALF_MAX_SIGNED(type) ((type)1 << (sizeof(type)*8-2))
#define __MAX_SIGNED(type) (__HALF_MAX_SIGNED(type) - 1 + __HALF_MAX_SIGNED(type))
#define __MIN_SIGNED(type) (-1 - __MAX_SIGNED(type))

#define __MIN(type) ((type)-1 < 1?__MIN_SIGNED(type):(type)0)
#define __MAX(type) ((type)~__MIN(type))

#define assigncheck(dest,src) ({ \
  typeof(src) __x=(src); \
  typeof(dest) __y=__x; \
  (__x==__y && ((__x<1) == (__y<1))?(void)((dest)=__y),0:1); \
})

#define add_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MIN(typeof(c))-(__b)<=(__a))?assigncheck(c,__a+__b):1) : \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a+__b):1); \
})

#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})

int mult32(long a,long b,long* c) {
    long long x=(long long)a*b;
    if (x!=(long)x) return 0;
  *c=x;
  return 1;
}

long power(i,j)
long i,j;
{
    long p = 1;
    long product;
    while (j-- > 0){
        assert0(mult32 (p, i, &product) == 1,"integer overflow in power" );
        p = product;
    }
    return (p);
}

#define int_plus(X,Y) int_node (node_ival(X) + node_ival(Y))
#define float_plus(X,Y) float_node(node_flval(X) + node_flval(Y))
#define int_minus(X,Y) int_node (node_ival(X) - node_ival(Y))
#define float_minus(X,Y) float_node(node_flval(X) - node_flval(Y))
#define int_mult(X,Y) int_node (node_ival(X) * node_ival(Y))
#define float_mult(X,Y) float_node(node_flval(X) * node_flval(Y))
#define int_div(X,Y) int_node (node_ival(X) / node_ival(Y))
#define float_div(X,Y) float_node(node_flval(X) / node_flval(Y))
#define int_mod(X,Y) int_node (mod (node_ival(X), node_ival(Y)))
#define float_mod(X,Y) float_node(fmod(node_flval(X),node_flval(Y)))
#define float_atantwo(X,Y) float_node(atan2(node_flval(X),node_flval(Y)))
#define int_exp(X,Y) int_node (power (node_ival(X), node_ival(Y)))
#define float_flpower(X,Y) float_node(pow(node_flval(X),node_flval(Y)))

/*  calculate value */
/* e1 and e2, e1 or e2 */
/* e1 < e2, e1 <= e2, e1 >= e2, e1 > e2 */ 
/* e1 + e2, e1 - e2, e1 * e2, e1 div e2, e1 mod e2, e1 ** e2 */
NODEPTR calculate_val(op, e1, e2)
unsigned op;
NODEPTR e1, e2;
{
    unsigned type = node_type(e1);
    long val;
    long product;
        /*double fl_product;*/

    
    
      type_check (e2, type);
      switch (op) {
      case AND:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e2 : e1);
	default:
	  break;
	}
	break;
      case OR:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e1 : e2);
	default:
	  break;
	}
	break;
      case LT:
          return (compare_val (e1, e2) < 0 ? True : False);
      case LE:
          return (compare_val (e1, e2) <= 0 ? True : False);
      case GE:
          return (compare_val (e1, e2) >= 0 ? True : False);
      case GT:
          return (compare_val (e1, e2) > 0 ? True : False);
      case PLUS:
	switch (type) {
	case INT:
            assert0(add_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in addition");
	  return (int_node (val));
        case FLOAT:
            return (float_plus (e1,e2));
	case LIST:
            return (append_list (e1, e2));
	case STRING:
            return (append_string (e1, e2));
	default:
	  break;
	}
	break;
      case MINUS:
	switch (type) {
	case INT:
            assert0(sub_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in subtraction");
	  return (int_node (val));
        case FLOAT:
            return (float_minus (e1,e2));  
	default:
	  break;
	}
	break;
      case MULT:
	switch (type) {
	case INT:
            assert0(mult32 (node_ival(e1), node_ival(e2), &product) == 1,"integer overflow in multiplication" );
            return (int_node(product));
        case FLOAT:
            return (float_mult (e1,e2));    
	default:
	  break;
	}
	break;  
      case DIV:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n"); 
	  return (int_div (e1, e2));
        case FLOAT:
            assert0(node_flval(e2) !=0, "division by zero is undefined\n");
            return (float_div (e1,e2));   
	default:
	  break;
	}
	break;
      case MOD:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n");
	  return (int_mod (e1, e2));
        case FLOAT:
            assert0(node_flval(e2) !=0, "division by zero is undefined\n");
	  return (float_mod (e1, e2));  
	default:
	  break;
	}
	break;
      case ATANTWO:
	switch (type) {
        case FLOAT:
	  return (float_atantwo (e1, e2));  
	default:
	  break;
	}
	break;  
      case POWER:
	switch (type) {
	case INT:
            assert0(node_ival(e2) >=0, "power to should be >=0\n");
            product=power(node_ival(e1), node_ival(e2));
	  return (int_node (product));
        case FLOAT:
            assert0(node_flval(e1)!=0 || node_flval(e2) >0, "domain error power to\n");
            assert0(node_flval(e1)>=0, "domain error power to\n");
            return (float_flpower(e1, e2));
	default:
	  break;
	}
	break;
      default:
	break;
      }
      error_break2 ("\"%s\" is not a %s operator",
		    Keyword[op], Keyword[type]);
      return (False);
}


/* Binary expressions: AND, OR, PLUS, MINUS, ... */
/* boolean expressions are speeded up otherwise use calculate_val*/
void reduce_bin_op(e)
NODEPTR e;
{
    NODEPTR calculate_val ();
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e)) {
            case AND:
                e2 = (node_bval(e1) == F) ? e1 : reduce_val (node_lnk_2(e));
                break;
            case OR:
                e2 = (node_bval(e1) == T) ? e1 : reduce_val (node_lnk_2(e));
                break;
            default:
                e2 = reduce_val (node_lnk_2(e));
        }
    }
    if (Ready)
        result (calculate_val (node_type(e), e1, e2));
}

#define int_uminus(X) int_node (- node_ival(X))
#define int_not(X) (node_ival(X) == 0 ? One : Zero)
#define float_uminus(X) float_node (- node_flval(X))
/* unary operators: NOT, UPLUS, UMINUS, SIZE, ... */
/* not e1, +e1, -e1, |e1| */
void reduce_u_op(e)
NODEPTR e;
{
    NODEPTR e1;

    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e1)) {
            case BOOL:
                switch (node_type(e)) {
                    case NOT:
                        result (node_bval(e1) ? False : True);
                        break;
                    default:
                        error_break1 ("\"%s\" is not a Boolean operator",
                                print_name (e));
                }
                break;
            case INT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (int_uminus (e1));
                        break;
                    case ITOF:
                        result (float_node((double)node_ival(e1)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not an integer operator",
                                   print_name (e));
                }
                break;
             case FLOAT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (float_uminus (e1));
                        break;
                    case CEIL:
                        result (int_node(ceil(node_flval(e1))));
                        break;
                    case FLOOR:
                        result (int_node(floor(node_flval(e1))));
                        break;
                    case SQRT:
                        result (float_node(sqrt(node_flval(e1))));
                        break;
                    case EXPO:
                        result (float_node(exp(node_flval(e1))));
                        break;
                    case LOG:
                        assert0(node_flval(e1)>0, "log argument should be greater than zero\n");
                        result (float_node(log(node_flval(e1))));
                        break;
                    case LOGTEN:
                        assert0(node_flval(e1)>0, "log10 argument should be greater than zero\n");
                        result (float_node(log10(node_flval(e1))));
                        break;
                    case SIN:
                        result (float_node(sin(node_flval(e1))));
                        break;
                    case COS:
                        result (float_node(cos(node_flval(e1))));
                        break;
                    case TAN:
                        result (float_node(tan(node_flval(e1))));
                        break;
                    case ASIN:
                        assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "asin argument should be in range [-1,1]\n");
                        result (float_node(asin(node_flval(e1))));
                        break;
                    case ACOS:
                        assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "acos argument should be in range [-1,1]\n");
                        result (float_node(acos(node_flval(e1))));
                        break;
                    case ATAN:
                        result (float_node(atan(node_flval(e1))));
                        break;
                    case SINH:
                        result (float_node(sinh(node_flval(e1))));
                        break;
                    case COSH:
                        result (float_node(cosh(node_flval(e1))));
                        break;
                    case TANH:
                        result (float_node(tanh(node_flval(e1))));
                        break;
                    case FABS:
                        result (float_node(fabs(node_flval(e1))));
                        break;    
                    default:
                        error_break1 ("\"%s\" is not an floating point operator",
                                   print_name (e));
                }
                break;   
            case STRING:
                switch (node_type(e)) {
                    case SIZE:
                        result (int_node ((int)node_lnk_2(e)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;
            default:
                error_break2 ("\"%s\" is not a %s operator",
                        print_name (e), print_name (e1));
        }
    }
}

/* random  */
/* Reduces to random integer >=0  AC 29Apr1999 */
void reduce_RANDOM()
{
  if ( RandomValTime != Clock || RandomVal == -1 ) {
    RandomVal = i4_uniform_0i (&TRandSeed);
    RandomValTime = Clock;
  };
  Ready = T;
  result (int_node (RandomVal));
}

/* Random  */
/* Reduces to random integer >=0  AC 20/09/2000 */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_RRANDOM()
{
  if ( RRandomValTime != Clock || RRandomVal == -1 ) {
    RRandomValTime = Clock;
  };
  Ready = T;
  result (int_node (i4_uniform_0i (&TRRandSeed)));
}

/* frandom  */
/* Reduces to random in [0.0,1.0) */
void reduce_SRANDOM()
{
  if ( SRandomValTime != Clock || SRandomVal == -1 ) {
    SRandomVal = r4_uniform_01 (&TSRandSeed);
    SRandomValTime = Clock;
  };
  Ready = T;
  result (float_node (SRandomVal));
}

/* fRandom  */
/* Reduces to random in [0.0,1.0) */
/* This will  allow for more than 1 frandom number
   in the same state */
void reduce_SRRANDOM()
{
  if ( SRRandomValTime != Clock || SRRandomVal == -1 ) {
    SRRandomValTime = Clock;
  };
  Ready = T;
  result (float_node (r4_uniform_01 (&TSRRandSeed)));
}

/********************\
*                    *
*  string operators  *
*                    *
\********************/

/* ascii(e1) */
void reduce_ASCII(e)
NODEPTR e;
{
    char   *strr;
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, STRING);
        assert1 (strlen (strr = node_strval(e1)) == 1,
                "cannot get ascii code of string \"%s\" (length not 1)", strr);
        result (int_node (*strr));
    }
}



/**********************\
*                      *
*  pointer operations  *
*                      *
\**********************/
/* ref(loc) */
void reduce_REFER(e)
NODEPTR e;
{
    NODEPTR loc = reduce_ref (node_lnk_1(e));
    if (Ready) {
        assert0 (loc != TempLoc, "cannot reference a value");
        result (loc_node (loc, is_next(loc))); 
    }
}
/* deref p */
void reduce_DEREF(e)
NODEPTR e;
{
    NODEPTR p = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (p, LOC); 
        ref_result (node_loc_tonext(p) ? next_loc (node_loc_ptr(p)) : curr_loc (node_loc_ptr(p)));
    }
}

